"""Copy records from the first KGTK file to the output file,
compacting repeated items into | lists.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkcompact import KgtkCompact
from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Copy a KGTK file compacting | lists.',
        'description': 'Copy a KGTK file, compacting multiple records into | lists. ' +
        '\n\nBy default, the input file is sorted in memory to achieve the ' +
        'grouping necessary for the compaction algorithm. This may cause ' +
        ' memory usage issues for large input files. If the input file has ' +
        'already been sorted (or at least grouped), the `--presorted` ' +
        'option may be used.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert compact --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_kgtk_file", nargs="?", type=Path, default="-",
                              help="The KGTK file to filter. May be omitted or '-' for stdin (default=%(default)s).")

    parser.add_argument(      "--columns", dest="key_column_names",
                              help="The key columns to identify records for compaction. " +
                              "(default=id for node files, (node1, label, node2, id) for edge files).", nargs='+', default=[ ])

    parser.add_argument(      "--compact-id", dest="compact_id",
                              help="Indicate that the ID column in KGTK edge files should be compacted. " +
                              "Normally, if the ID column exists, it is not compacted, " +
                              "as there are use cases that need to maintain distinct lists of secondary edges for each ID value. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--presorted", dest="sorted_input",
                              help="Indicate that the input has been presorted (or at least pregrouped) (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--verify-sort", dest="verify_sort",
                              help="If the input has been presorted, verify its consistency (disable if only pregrouped). (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--build-id", dest="build_id",
                              help="Build id values in an id column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    
    KgtkIdBuilderOptions.add_arguments(parser, expert=_expert)
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],
        key_column_names: typing.List[str],
        compact_id: bool,
        sorted_input: bool,
        verify_sort: bool,
        build_id: bool,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_dict(kwargs)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--columns=%s" % " ".join(key_column_names), file=error_file)
        print("--compact-id=%s" % str(compact_id), file=error_file, flush=True)
        print("--presorted=%s" % str(sorted_input))
        print("--verify-sort=%s" % str(verify_sort), file=error_file, flush=True)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        print("--build-id=%s" % str(build_id), file=error_file, flush=True)
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ex: KgtkCompact = KgtkCompact(
            input_file_path=input_kgtk_file,
            key_column_names=key_column_names,
            compact_id=compact_id,
            sorted_input=sorted_input,
            verify_sort=verify_sort,
            output_file_path=output_kgtk_file,
            build_id=build_id,
            idbuilder_options=idbuilder_options,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ex.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Merge identical nodes and deduplicate.
"""


import sys

def parser():
    return {
        'help': 'Merge identical nodes and deduplicate.'
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    parser.add_argument( "-nf", "--node-file", action="store", type=str, dest="nodes_file", help="TSV file with node columns.")
    parser.add_argument( "-ef", "--edge-file", action="store", type=str, dest="edges_file", help="TSV file with edge columns.")
    parser.add_argument("-l", "--label", action="store", type=str, dest="label", help="Relation label indicating identity")

def run(nodes_file, edges_file, label): 
    # import modules locally
    import socket
    import sh # type: ignore
    from kgtk.exceptions import KGTKException
    from kgtk.cskg_utils import collapse_identical_nodes
    try:
        new_edges_df, new_nodes_df = collapse_identical_nodes(edges_file, nodes_file)
        print(new_edges_df)
        print(new_nodes_df)
    except:
        raise KGTKException



"""Filter a KGTK file based on whether one or more records exist in a second
KGTK file with matching values for one or more fields.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.iff.kgtkifempty import KgtkIfEmpty
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Filter a KGTK file for empty fields.',
        'description': 'Filter a KGTK file based on whether one or more fields are empty. ' +
        'When multiple fields are specified, either any field or all fields must be empty.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert ifempty --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    parser.add_argument(      "input_kgtk_file", nargs="?", help="The KGTK file to filter. May be omitted or '-' for stdin.", type=Path)

    parser.add_argument(      "--columns", dest="filter_column_names",
                              help="The columns in the file being filtered (Required).", nargs='+', required=True)

    parser.add_argument(      "--count", dest="only_count", help="Only count the records, do not copy them. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--all", dest="all_are", help="False: Test if any are empty, True: test if all are empty (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],
        filter_column_names: typing.List[str],
        all_are: bool = False,

        only_count: bool = False,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--columns=%s" % " ".join(filter_column_names), file=error_file)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        print("--count=%s" % str(only_count))
        print("--all=%s" % str(all_are))
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ie: KgtkIfEmpty = KgtkIfEmpty(
            input_file_path=input_kgtk_file,
            filter_column_names=filter_column_names,
            output_file_path=output_kgtk_file,
            all_are=all_are,
            notempty=False,
            only_count=only_count,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ie.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Validate a KGTK file, producing error messages.

At the present time, validation looks at such things as:
1)      Presence of require columns
2)      Consistent number of columns
3)      Comments, whitespace lines, line s with empty required columns

Certain constraints can be overlooked or repaired.

This program does not validate individual fields.
"""

from argparse import Namespace
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Validate a KGTK file ',
        
        'description': 'Validate a KGTK file. ' +
        'Empty lines, whitespace lines, comment lines, and lines with empty required fields are silently skipped. ' +
        'Header errors cause an immediate exception. Data value errors are reported. ' +

        '\n\nTo validate data and pass clean data to an output file or pipe, use the kgtk clean_data command.' +

        '\n\nAdditional options are shown in expert help.\nkgtk --expert validate --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    _expert: bool = parsed_shared_args._expert

    parser.add_argument(      "kgtk_files", nargs="*", help="The KGTK file(s) to validate. May be omitted or '-' for stdin.", type=Path)

    parser.add_argument(      "--header-only", dest="header_only",
                              help="Process the only the header of the input file (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, validate_by_default=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(kgtk_files: typing.Optional[typing.List[typing.Optional[Path]]],
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = False,
        header_only: bool = False,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,
        **kwargs # Whatever KgtkReaderOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    if kgtk_files is None or len(kgtk_files) == 0:
        kgtk_files = [ None ]

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stderr if errors_to_stderr else sys.stdout

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % " ".join((str(kgtk_file) for kgtk_file in kgtk_files)), file=error_file)
        print("--header-only=%s" % str(header_only), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kgtk_file: typing.Optional[Path]
        for kgtk_file in kgtk_files:
            if verbose:
                print("\n====================================================", flush=True)
                if kgtk_file is not None:
                    print("Validating '%s'" % str(kgtk_file), file=error_file, flush=True)
                else:
                    print ("Validating from stdin", file=error_file, flush=True)

            kr: KgtkReader = KgtkReader.open(kgtk_file,
                                             error_file=error_file,
                                             options=reader_options,
                                             value_options=value_options,
                                             verbose=verbose,
                                             very_verbose=very_verbose)
        
            if header_only:
                kr.close()
                if verbose:
                    print("Validated the header only.", file=error_file, flush=True)
            else:
                line_count: int = 0
                row: typing.List[str]
                for row in kr:
                    line_count += 1
                if verbose:
                    print("Validated %d data lines" % line_count, file=error_file, flush=True)
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""Copy records from the first ("left") KGTK file to the output file, if
a match is made with records in the second ("right") KGTK input file.

The fields to match may be supplied by the user.  If not supplied,
the following defaults will be used:

Left    Right   Key fields
edge    edge    left.node1 = right.node1 and left.label=right.label and left.node2=right.node2
node    node    left.id = right.id
edge    node    left.node1 = right.id
node    edge    right.id = left.node1

Note: By default, this implementation builds im-memory sets of all the key
values in the second file (the filter file). Optionally, it will cache the
first file (the input file) instead.

Note: By default, input records are passed in order to the output file.  When
the input file is cached, the output records are order by key value (alpha
sort), then by input order.  However, --preserve-order can be used to retain
the input file's order in the output file.

TODO: Study the time and space tradeoff between process_cacheing_input(...)
and process_cacheing_input_preserving_order(...).  Perhaps there's no reason
for both algorithms?
"""

from argparse import ArgumentParser, Namespace
import attr
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

@attr.s(slots=True, frozen=True)
class KgtkIfExists(KgtkFormat):
    input_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))
    input_keys: typing.Optional[typing.List[str]] = attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                                                            iterable_validator=attr.validators.instance_of(list))))

    filter_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    filter_keys: typing.Optional[typing.List[str]] = attr.ib(validator=attr.validators.optional(attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str),
                                                                                                                             iterable_validator=attr.validators.instance_of(list))))

    output_file_path: typing.Optional[Path] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(Path)))

    # The field separator used in multifield joins.  The KGHT list character should be safe.
    field_separator: str = attr.ib(validator=attr.validators.instance_of(str), default=KgtkFormat.LIST_SEPARATOR)

    invert: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    cache_input: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    preserve_order: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    # TODO: find working validators
    # value_options: typing.Optional[KgtkValueOptions] = attr.ib(attr.validators.optional(attr.validators.instance_of(KgtkValueOptions)), default=None)
    input_reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    filter_reader_options: typing.Optional[KgtkReaderOptions]= attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    FIELD_SEPARATOR_DEFAULT: str = KgtkFormat.LIST_SEPARATOR
    
    def get_primary_key_column(self, kr: KgtkReader, who: str)->typing.List[int]:
        if kr.is_node_file:
            if kr.id_column_idx < 0:
                raise ValueError("The id column is missing from the %s node file." % who)
            return [ kr.id_column_idx ]
        elif kr.is_edge_file:
            if kr.node1_column_idx < 0:
                raise ValueError("The node1 column is missing from the %s node file." % who)
            return [ kr.node1_column_idx ]
        else:
            raise ValueError("The %s file is neither edge nor node." % who)

    def get_edge_key_columns(self, kr: KgtkReader, who: str)-> typing.List[int]:
        if not kr.is_edge_file:
            raise ValueError("get_edge_keys called on %s at wrong time." % who)
        if kr.node1_column_idx < 0:
            raise ValueError("The node1 column is missing from the %s edge file." % who)
        if kr.label_column_idx < 0:
            raise ValueError("The label column is missing from the %s edge file." % who)
        if kr.node2_column_idx < 0:
            raise ValueError("The node2 column is missing from the %s edge file." % who)
        return [ kr.node1_column_idx, kr.label_column_idx, kr.node2_column_idx ]

    def get_supplied_key_columns(self, supplied_keys: typing.List[str], kr: KgtkReader, who: str)->typing.List[int]:
        result: typing.List[int] = [ ]
        key: str
        for key in supplied_keys:
            if key not in kr.column_name_map:
                raise ValueError("Column %s is not in the %s file" % (key, who))
            result.append(kr.column_name_map[key])
        return result
    
    def get_key_columns(self, supplied_keys: typing.Optional[typing.List[str]], kr: KgtkReader, other_kr: KgtkReader, who: str)->typing.List[int]:
        if supplied_keys is not None and len(supplied_keys) > 0:
            return self.get_supplied_key_columns(supplied_keys, kr, who)

        if not (kr.is_node_file or kr.is_edge_file):
            raise ValueError("The %s file is a quasi-KGTK file.  Please supply its keys." % who)

        if kr.is_node_file or other_kr.is_node_file:
            return self.get_primary_key_column(kr, who)

        return self.get_edge_key_columns(kr, who)

    def build_key(self, row: typing.List[str], key_columns: typing.List[int])->str:
        key: str = ""
        idx: int
        first: bool = True
        for idx in key_columns:
            if first:
                first = False
            else:
                key += self.field_separator
            key += row[idx]
        return key

    def extract_key_set(self, kr: KgtkReader, who: str, key_columns: typing.List[int])->typing.Set[str]:
        key_set: typing.Set[str] = set()
        row: typing.List[str]
        for row in kr:
            key_set.add(self.build_key(row, key_columns))
        return key_set

    def extract_key_set_and_cache(self, kr: KgtkReader, who: str, key_columns: typing.List[int])->typing.Tuple[typing.Set[str], typing.List[typing.List[str]]]:
        key_set: typing.Set[str] = set()
        cache: typing.List[typing.List[str]] = [ ]
        row: typing.List[str]
        for row in kr:
            key_set.add(self.build_key(row, key_columns))
            cache.append(row)
        return key_set, cache

    def process_cacheing_filter(self,
                                input_kr: KgtkReader,
                                filter_kr: KgtkReader,
                                input_key_columns: typing.List[int],
                                filter_key_columns: typing.List[int],
                                ew: KgtkWriter):
        if self.verbose:
            print("Processing by cacheing the filter file's key set..")

        if self.verbose:
            print("Building the filter key set from %s" % self.filter_file_path, file=self.error_file, flush=True)
        key_set: typing.Set[str] = self.extract_key_set(filter_kr, "filter", filter_key_columns)
        if self.verbose or self.very_verbose:
            print("There are %d entries in the filter key set." % len(key_set), file=self.error_file, flush=True)
            if self.very_verbose:
                print("Keys: %s" % " ".join(key_set), file=self.error_file, flush=True)

        if self.verbose:
            print("Filtering records from %s" % self.input_file_path, file=self.error_file, flush=True)
        input_line_count: int = 0
        output_line_count: int = 0;

        # TODO: join these two code paths using xor?
        row: typing.List[str]
        input_key: str
        if self.invert:
            for row in input_kr:
                input_line_count += 1
                input_key = self.build_key(row, input_key_columns)
                if input_key not in key_set:
                    ew.write(row)
                    output_line_count += 1
        else:
            for row in input_kr:
                input_line_count += 1
                input_key = self.build_key(row, input_key_columns)
                if input_key in key_set:
                    ew.write(row)
                    output_line_count += 1

        if self.verbose:
            print("Read %d records, wrote %d records." % (input_line_count, output_line_count), file=self.error_file, flush=True)
        

    def process_cacheing_input(self,
                               input_kr: KgtkReader,
                               filter_kr: KgtkReader,
                               input_key_columns: typing.List[int],
                               filter_key_columns: typing.List[int],
                               ew: KgtkWriter):
        if self.verbose:
            print("Processing by cacheing the input file.")
        input_line_count: int = 0
        filter_line_count: int = 0
        output_line_count: int = 0
        
        # Map key values to lists of input and output data.
        inputmap: typing.MutableMapping[str, typing.List[typing.List[str]]] = { }
        outputmap: typing.MutableMapping[str, typing.List[typing.List[str]]] = { }

        if self.verbose:
            print("Reading the input data from %s" % self.input_file_path, file=self.error_file, flush=True)
        row: typing.List[str]
        for row in input_kr:
            input_line_count += 1
            input_key: str = self.build_key(row, input_key_columns)
            if input_key in inputmap:
                # Append the row to an existing list for that key.
                inputmap[input_key].append(row)
            else:
                # Create a new list of rows for this key.
                inputmap[input_key] = [ row ]

        if self.verbose:
            print("Applying the filter from %s" % self.filter_file_path, file=self.error_file, flush=True)
        filter_key: str
        if self.invert:
            outputmap = inputmap
            for row in filter_kr:
                filter_line_count += 1
                filter_key = self.build_key(row, filter_key_columns)
                if filter_key in outputmap:
                    del outputmap[filter_key]
        else:
            for row in filter_kr:
                filter_line_count += 1
                filter_key = self.build_key(row, filter_key_columns)
                if filter_key in inputmap:
                    outputmap[filter_key] = inputmap[filter_key]

        if self.verbose:
            print("Writing the output data to %s" % self.output_file_path, file=self.error_file, flush=True)

        # To simplify debugging, write the output data in sorted order (keys,
        # then input order).
        key: str
        for key in sorted(outputmap.keys()):
            for row in outputmap[key]:
                ew.write(row)
                output_line_count += 1

        if self.verbose:
            print("Read %d input records, read %d filter records, wrote %d records." % (input_line_count,
                                                                                        filter_line_count,
                                                                                        output_line_count),
                  file=self.error_file, flush=True)

    def process_cacheing_input_preserving_order(self,
                                                input_kr: KgtkReader,
                                                filter_kr: KgtkReader,
                                                input_key_columns: typing.List[int],
                                                filter_key_columns: typing.List[int],
                                                ew: KgtkWriter):
        # This algorithm preserves the input file's record order in the output file,
        # at the cost of extra work building keys.

        if self.verbose:
            print("Processing by cacheing the input file while preserving record order.")

        # Step one:  read the input file, cache it, and build the input key set
        if self.verbose:
            print("Building the input key set from %s" % self.input_file_path, file=self.error_file, flush=True)
        input_key_set: typing.Set[str]
        input_cache: typing.List[typing.List[str]]
        input_key_set, input_cache = self.extract_key_set_and_cache(input_kr, "input", input_key_columns)
        input_line_count: int = len(input_cache)
        if self.verbose or self.very_verbose:
            print("There are %d rows in the input cache." % input_line_count, file=self.error_file, flush=True)
            print("There are %d entries in the input key set." % len(input_key_set), file=self.error_file, flush=True)
            if self.very_verbose:
                print("Keys: %s" % " ".join(input_key_set), file=self.error_file, flush=True)

        # Step two: read the filter file and act on the key_set.
        output_key_set: typing.Set[str] = set()
        if self.verbose:
            print("Applying the filter from %s" % self.filter_file_path, file=self.error_file, flush=True)
        filter_key: str
        filter_line_count: int = 0
        row: typing.List[str]
        if self.invert:
            output_key_set = input_key_set
            for row in filter_kr:
                filter_line_count += 1
                filter_key = self.build_key(row, filter_key_columns)
                if filter_key in output_key_set:
                    output_key_set.remove(filter_key)
        else:
            for row in filter_kr:
                filter_line_count += 1
                filter_key = self.build_key(row, filter_key_columns)
                if filter_key in input_key_set:
                    output_key_set.add(filter_key)
        if self.verbose:
            print("Read %d rows from the filter file." % filter_line_count, file=self.error_file, flush=True)
            print("There are %d entries in the output key set." % len(output_key_set), file=self.error_file, flush=True)

        # Step three: read the input rows from the cache and write only the
        # ones with keys in the output key set.
        output_line_count: int = 0
        for row in input_cache:
            input_key: str = self.build_key(row, input_key_columns)
            if input_key in output_key_set:
                ew.write(row)
                output_line_count += 1
        if self.verbose:
            print("Wrote %d rows to the output file." % output_line_count, file=self.error_file, flush=True)
        
    def process(self):
        # Open the input files once.
        if self.verbose:
            if self.input_file_path is not None:
                print("Opening the input file: %s" % self.input_file_path, file=self.error_file, flush=True)
            else:
                print("Reading the input data from stdin", file=self.error_file, flush=True)

        input_kr: KgtkReader =  KgtkReader.open(self.input_file_path,
                                                error_file=self.error_file,
                                                who="input",
                                                options=self.input_reader_options,
                                                value_options = self.value_options,
                                                verbose=self.verbose,
                                                very_verbose=self.very_verbose,
        )

        if self.verbose:
            print("Opening the filter input file: %s" % self.filter_file_path, flush=True)
        filter_kr: KgtkReader = KgtkReader.open(self.filter_file_path,
                                                who="filter",
                                                error_file=self.error_file,
                                                options=self.filter_reader_options,
                                                value_options=self.value_options,
                                                verbose=self.verbose,
                                                very_verbose=self.very_verbose,
        )

        input_key_columns: typing.List[int] = self.get_key_columns(self.input_keys, input_kr, filter_kr, "input")
        filter_key_columns: typing.List[int] = self.get_key_columns(self.filter_keys, filter_kr, input_kr, "filter")

        if len(input_key_columns) != len(filter_key_columns):
            print("There are %d input key columns but %d filter key columns.  Exiting." % (len(input_key_columns), len(filter_key_columns)),
                  file=self.error_file, flush=True)
            return

        if self.verbose:
            print("Opening the output file: %s" % self.output_file_path, file=self.error_file, flush=True)
        ew: KgtkWriter = KgtkWriter.open(input_kr.column_names,
                                         self.output_file_path,
                                         mode=input_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)

        if self.cache_input:
            if self.preserve_order:
                pass
                self.process_cacheing_input_preserving_order(input_kr=input_kr,
                                                             filter_kr=filter_kr,
                                                             input_key_columns=input_key_columns,
                                                             filter_key_columns=filter_key_columns,
                                                             ew=ew)
            else:
                self.process_cacheing_input(input_kr=input_kr,
                                            filter_kr=filter_kr,
                                            input_key_columns=input_key_columns,
                                            filter_key_columns=filter_key_columns,
                                            ew=ew)
        else:
            self.process_cacheing_filter(input_kr=input_kr,
                                         filter_kr=filter_kr,
                                         input_key_columns=input_key_columns,
                                         filter_key_columns=filter_key_columns,
                                         ew=ew)

        ew.close()

def main():
    """
    Test the KGTK file joiner.
    """
    parser: ArgumentParser = ArgumentParser()

    parser.add_argument(dest="input_file_path", help="The KGTK file with the input data", type=Path, nargs="?")

    parser.add_argument(      "--filter-on", dest="filter_file_path", help="The KGTK file with the filter data (required).", type=Path, required=True)

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")
    
    parser.add_argument(      "--field-separator", dest="field_separator", help="Separator for multifield keys (default=%(default)s)",
                              default=KgtkIfExists.FIELD_SEPARATOR_DEFAULT)
   
    parser.add_argument(      "--invert", dest="invert", help="Invert the test (if not exists) (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--cache-input", dest="cache_input", help="Cache the input file instead of the filter keys. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--preserve-order", dest="preserve_order", help="Preserve record order when cacheing the input file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--input-keys", dest="input_keys", help="The key columns in the input file (default=None).", nargs='*')
    parser.add_argument(      "--filter-keys", dest="filter_keys", help="The key columns in the filter file (default=None).", nargs='*')

    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input")
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="filter")
    KgtkValueOptions.add_arguments(parser)

    args: Namespace = parser.parse_args()

    error_file: typing.TextIO = sys.stdout if args.errors_to_stdout else sys.stderr

    # Build the option structures.                                                                                                                          
    input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args, who="input")
    filter_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_args(args, who="filter")
    value_options: KgtkValueOptions = KgtkValueOptions.from_args(args)

   # Show the final option structures for debugging and documentation.                                                                                             
    if args.show_options:
        print("input: %s" % (str(args.input_file_path) if args.input_file_path is not None else "-"), file=error_file)
        print("--filter-on=%s" % str(args.filter_file_path), file=error_file)
        print("--output-file=%s" % str(args.output_file_path), file=error_file)
        print("--field-separator=%s" % repr(args.field_separator), file=error_file)
        print("--invert=%s" % str(args.invert), file=error_file)
        print("--cache-input=%s" % str(args.cache_input), file=error_file)
        print("--preserve-order=%s" % str(args.preserve_order), file=error_file)
        if args.input_keys is not None:
            print("--input-keys %s" % " ".join(args.input_keys), file=error_file)
        if args.filter_keys is not None:
            print("--filter-keys %s" % " ".join(args.filter_keys), file=error_file)
        input_reader_options.show(out=error_file, who="input")
        filter_reader_options.show(out=error_file, who="filter")
        value_options.show(out=error_file)

    ie: KgtkIfExists = KgtkIfExists(
        input_file_path=args.input_file_path,
        input_keys=args.input_keys,
        filter_file_path=args.filter_file_path,
        filter_keys=args.filter_keys,
        output_file_path=args.output_file_path,
        field_separator=args.field_separator,
        invert=args.invert,
        cache_input=args.cache_input,
        preserve_order=args.preserve_order,
        input_reader_options=input_reader_options,
        filter_reader_options=filter_reader_options,
        value_options=value_options,
        error_file=error_file,
        verbose=args.verbose,
        very_verbose=args.very_verbose)

    ie.process()

if __name__ == "__main__":
    main()



"""Copy records from the first KGTK file to the output file,
expanding | lists.

TODO: Need KgtkWriterOptions
"""

from argparse import _MutuallyExclusiveGroup, Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.kgtkformat import KgtkFormat
from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkexplode import KgtkExplode
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalue import KgtkValueFields
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'aliases': ['denormalize_node2'],
        'help': 'Copy a KGTK file, exploding one column (usualy node2) into seperate columns for each subfield.',
        'description': 'Copy a KGTK file, exploding one column (usually node2) into seperate columns for each subfield. ' +
        'If a cell in the column being exploded contains a list, that record is optionally expanded into multiple records ' +
        'before explosion, with all other columns copied-as is.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert expand --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_kgtk_file", nargs="?", type=Path, default="-",
                              help="The KGTK file to filter. May be omitted or '-' for stdin (default=%(default)s).")

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--column", dest="column_name", help="The name of the column to explode. (default=%(default)s).", default=KgtkFormat.NODE2)

    fgroup: _MutuallyExclusiveGroup = parser.add_mutually_exclusive_group()
    fgroup.add_argument(      "--types", dest="type_names", nargs='*',
                              help="The KGTK data types for which fields should be exploded. (default=%(default)s).",
                              choices=KgtkFormat.DataType.choices(),
                              default=KgtkFormat.DataType.choices())

    fgroup.add_argument(      "--fields", dest="field_names", nargs='*', 
                              help=h("The names of the fields to extract (overrides --types). (default=%(default)s)."),
                              choices=KgtkValueFields.FIELD_NAMES)

    parser.add_argument(      "--prefix", dest="prefix", help="The prefix for exploded column names. (default=%(default)s).",
                              default=KgtkFormat.NODE2 + ";" + KgtkFormat.KGTK_NAMESPACE)

    parser.add_argument(      "--overwrite", dest="overwrite_columns",
                              help="Indicate that it is OK to overwrite existing columns. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--expand", dest="expand_list",
                              help="Expand the source column if it contains a list, else fail. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--show-data-types", dest="show_data_types",
                              help="Print the list of data types and exit. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],
        column_name: str,
        type_names: typing.List[str],
        field_names: typing.List[str],
        prefix: str,
        overwrite_columns: bool,
        expand_list: bool,
        show_data_types: bool,
        
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--column %s" % column_name, file=error_file, flush=True)
        print("--prefix %s" % prefix, file=error_file, flush=True)
        print("--overwrite %s" % str(overwrite_columns), file=error_file, flush=True)
        print("--expand %s" % str(expand_list), file=error_file, flush=True)
        if type_names is not None:
            print("--types %s" % " ".join(type_names), file=error_file, flush=True)
        if field_names is not None:
            print("--fields %s" % " ".join(field_names), file=error_file, flush=True)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        print("--show-data-types %s" % str(show_data_types), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)
    if show_data_types:
        data_type: str
        for data_type in KgtkFormat.DataType.choices():
            print("%s" % data_type, file=error_file, flush=True)
        return 0

    try:
        ex: KgtkExplode = KgtkExplode(
            input_file_path=input_kgtk_file,
            output_file_path=output_kgtk_file,
            column_name=column_name,
            type_names=type_names,
            field_names=field_names,
            prefix=prefix,
            overwrite_columns=overwrite_columns,
            expand_list=expand_list,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ex.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Remove columns from a file.
"""


import sys


def parser():
    return {
        'help': 'Remove columns from a file '
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    parser.add_argument('-dt', "--datatype", action="store", type=str, dest="datatype",
                        help="Datatype of the input file, e.g., tsv or csv.", default="tsv")
    parser.add_argument('-c', "--columns", action="store", type=str, dest="columns",
                        help="Columns to remove as a comma-separated string, e.g., id,docid")
    parser.add_argument('input', nargs='?', action='store')


def run(datatype, columns, input): 
	# import modules locally
	import socket
	import sh # type: ignore
	from kgtk.exceptions import KGTKException

	try:
		if input:

			print(input)
			sh.mlr('--tsvlite', 'cut', '-x', '-f', columns, 
					input, _out=sys.stdout, _err=sys.stderr)
		elif not sys.stdin.isatty():
			print(sh.mlr('--tsvlite', 'cut', '-x', '-f', columns,
						 _in=sys.stdin, _out=sys.stdout, _err=sys.stderr))
		else:
			raise KGTKException
	except:
		raise KGTKException
    #parser.print_help()




from argparse import ArgumentParser, RawDescriptionHelpFormatter
from functools import partial
from kgtk.exceptions import KGTKArgumentParseException, KGTKSyntaxException


class KGTKArgumentParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        if not kwargs.get('formatter_class'):
            kwargs['formatter_class'] = RawDescriptionHelpFormatter

        self.shared_arguments = set()
        self.default_arguments = set()
        self.add_default_argument_funcs = {}

        super(KGTKArgumentParser, self).__init__(*args, **kwargs)

    def add_default_argument(self, *args, **kwargs):
        if 'dest' not in kwargs:
            raise KGTKSyntaxException('Default argument error: dest not defined')
        self.add_default_argument_funcs[kwargs['dest']] = partial(self.add_argument, *args, **kwargs)

    def accept_shared_argument(self, dest):
        self.shared_arguments.add(dest)

    def accept_default_argument(self, dest):
        # load default arguments
        self.add_default_argument_funcs[dest]()

    def exit(self, status=0, message=None):
        if status == 2:
            status = KGTKArgumentParseException.return_code
        super(KGTKArgumentParser, self).exit(status, message)


def add_shared_arguments(parser):
    # set shared arguments here
    # 1. no flag, only name (starts with two dashes: --XXX)
    # 2. need to have default value
    # 3. need to specify dest, which value starts with underscore
    # e.g., parser.add_argument('--debug', dest='_debug', action='store_true', default=False, help='enable debug mode')
    pass


def add_default_arguments(parser):
    # set default arguments here
    # need to specify dest
    # e.g., parser.add_default_argument('--save', dest='save', action='store', help='save to file')
    pass



"""Filter a KGTK file based on whether one or more records exist in a second
KGTK file with matching values for one or more fields.

Note: By default, this implementation builds im-memory sets of all the key
values in the second file (the filter file). Optionally, it will cache the
first file (the input file) instead.

Note: By default, input records are passed in order to the output file.  When
the input file is cached, the output records are order by key value (alpha
sort), then by input order.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.iff.kgtkifexists import KgtkIfExists
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Filter a KGTK file by not matching records in a second KGTK file.',
        'description': 'Filter a KGTK file based on whether one or more records do not exist in a second KGTK file with matching values for one or more fields.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert ifnotexists --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_kgtk_file", nargs="?", help="The KGTK file to filter. May be omitted or '-' for stdin.", type=Path)

    parser.add_argument(      "--input-keys", "--left-keys", dest="input_keys",
                              help="The key columns in the file being filtered (default=None).", nargs='*')

    parser.add_argument(      "--filter-on", dest="filter_kgtk_file", help="The KGTK file to filter against (required).", type=Path, required=True)

    parser.add_argument(      "--filter-keys", "--right-keys", dest="filter_keys",
                              help="The key columns in the filter-on file (default=None).", nargs='*')

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (required),", type=Path, default=None)

    parser.add_argument(      "--cache-input", dest="cache_input", help="Cache the input file instead of the filter keys (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--preserve-order", dest="preserve_order", help="Preserve record order when cacheing the input file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--field-separator", dest="field_separator",
                              help=h("Separator for multifield keys (default=%(default)s)"),
                              default=KgtkIfExists.FIELD_SEPARATOR_DEFAULT)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="filter", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        filter_kgtk_file: Path,
        output_kgtk_file: typing.Optional[Path],
        input_keys: typing.Optional[typing.List[str]],
        filter_keys: typing.Optional[typing.List[str]],
        
        cache_input: bool = False,
        preserve_order: bool = False,

        field_separator: str = KgtkIfExists.FIELD_SEPARATOR_DEFAULT,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException


    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="input", fallback=True)
    filter_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="filter", fallback=True)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        if input_keys is not None:
            print("--input-keys=%s" % " ".join(input_keys), file=error_file)
        print("--filter-on=%s" % (str(filter_kgtk_file) if filter_kgtk_file is not None else "-"), file=error_file)
        if filter_keys is not None:
            print("--filter-keys=%s" % " ".join(filter_keys), file=error_file)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        print("--cache-input=%s" % str(cache_input), file=error_file)
        print("--preserve-order=%s" % str(preserve_order), file=error_file)
        print("--field-separator='%s'" % repr(field_separator), file=error_file)
        input_reader_options.show(out=error_file, who="input")
        filter_reader_options.show(out=error_file, who="filter")
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ie: KgtkIfExists = KgtkIfExists(
            input_file_path=input_kgtk_file,
            input_keys=input_keys,
            filter_file_path=filter_kgtk_file,
            filter_keys=filter_keys,
            output_file_path=output_kgtk_file,
            invert=True,
            cache_input=cache_input,
            preserve_order=preserve_order,
            field_separator=field_separator,
            input_reader_options=input_reader_options,
            filter_reader_options=filter_reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ie.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions


def parser():
    return {
        'help': 'Find connected components in a Graph.',
        'description': 'Find all the connected components in an undirected or directed Graph.' +
                       '\n\nAdditional options are shown in expert help.\nkgtk --expert connected-components --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    _expert: bool = parsed_shared_args._expert

    parser.add_argument("input_kgtk_file", nargs="?",
                        help="The KGTK file to find connected components in. May be omitted or '-' for stdin.",
                        type=Path)
    parser.add_argument("--no-header", action="store_true", dest="no_header",
                        help="Specify if the input file does not have a header, default FALSE")
    parser.add_argument("--properties", action="store", type=str, dest="properties",
                        help=' A comma separated list of properties to traverse while finding connected components, '
                             'by default all properties will be considered',
                        default='')
    parser.add_argument('--undirected', action='store_true', dest="undirected",
                        help="Specify if the input graph is undirected, default FALSE")
    parser.add_argument('--strong', action='store_true', dest="strong",
                        help="Treat graph as directed or not, independent of its actual directionality.")
    parser.add_argument("-o", "--output-file", dest="output_kgtk_file",
                        help="The KGTK file to write, by default the output will be written to standard output).",
                        type=Path, default=None)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", expert=_expert, defaults=False)
    # KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],
        no_header: bool = False,
        properties: str = None,
        undirected: bool = False,
        strong: bool = False,
        **kwargs  # Whatever KgtkFileOptions and KgtkValueOptions want.
        ) -> int:
    from kgtk.gt.connected_components import ConnectedComponents
    from kgtk.exceptions import KGTKException
    cc: ConnectedComponents = ConnectedComponents(input_file_path=input_kgtk_file,
                                                  output_file_path=output_kgtk_file,
                                                  no_header=no_header,
                                                  properties=properties,
                                                  undirected=undirected,
                                                  strong=strong)

    try:
        cc.process()
    except Exception as e:
        raise KGTKException(str(e))



"""Copy records from the first KGTK file to the output file,
expanding | lists.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.reshape.kgtkexpand import KgtkExpand
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Copy a KGTK file expanding | lists.',
        'description': 'Copy a KGTK file, expanding | lists into multiple records. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert expand --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    parser.add_argument(      "input_kgtk_file", nargs="?", type=Path, default="-",
                              help="The KGTK file to filter. May be omitted or '-' for stdin (default=%(default)s).")

    parser.add_argument(      "--columns", dest="key_column_names",
                              help="The key columns will not be expanded.  They will be repeated on each output record. " +
                              "(default=id for node files, (node1, label, node2) for edge files).", nargs='+', default=[ ])


    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: typing.Optional[Path],
        output_kgtk_file: typing.Optional[Path],
        key_column_names: typing.List[str],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--columns=%s" % " ".join(key_column_names), file=error_file)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ex: KgtkExpand = KgtkExpand(
            input_file_path=input_kgtk_file,
            key_column_names=key_column_names,
            output_file_path=output_kgtk_file,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ex.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""Add label columns for values in the node1, label, and node2 fields.

The input rows are saved in memory, as well as the value-to-label mapping.
This will impose a limit on the size of the input files that can be processed.

TODO: Optionally save the input rows in an external disk file?

TODO: Optionally reread the input stream insted of saving the input rows?
      With special provision for copying standard input to an external disk file.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.lift.kgtklift import KgtkLift
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Lift labels from a KGTK file.',
        'description': 'Lift labels for a KGTK file. For each of the items in the (node1, label, node2) columns, look for matching label records. ' +
        'If found, lift the label values into additional columns in the current record. ' +
        'Label records are reoved from the output. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert lift --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "input_kgtk_file", nargs="?", help="The KGTK file to lift. May be omitted or '-' for stdin.", type=Path, default="-")

    parser.add_argument(      "--label-file", dest="label_kgtk_file", help="A KGTK file with label records (default=%(default)s).", type=Path, default=None)

    parser.add_argument(      "--node1-name", dest="node1_column_name",
                              help=h("The name of the node1 column. (default=node1 or alias)."), default=None)

    parser.add_argument(      "--label-name", dest="label_column_name",
                              help=h("The name of the label column. (default=label)."), default=None)

    parser.add_argument(      "--node2-name", dest="node2_column_name",
                              help=h("The name of the node2 column. (default=node2 or alias)."), default=None)

    parser.add_argument(      "--label-value", dest="label_column_value", help=h("The value in the label column. (default=%(default)s)."), default="label")
    parser.add_argument(      "--lift-suffix", dest="lifted_column_suffix",
                              help=h("The suffix used for newly created columns. (default=%(default)s)."), default=";label")

    parser.add_argument(      "--columns-to-lift", dest="lift_column_names", help=h("The columns to lift. (default=[node1, label, node2])."), nargs='*')

    parser.add_argument("-o", "--output-file", dest="output_kgtk_file", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--remove-label-records", dest="remove_label_records",
                              help=h("If true, remove label records from the output. (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--sort-lifted-labels", dest="sort_lifted_labels",
                              help=h("If true, sort lifted labels with lists. (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-duplicate-labels", dest="suppress_duplicate_labels",
                              help=h("If true, suppress duplicate values in lists in lifted labels (implies sorting). (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=True)

    parser.add_argument(      "--suppress-empty-columns", dest="suppress_empty_columns",
                              help="If true, do not create new columns that would be empty. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--ok-if-no-labels", dest="ok_if_no_labels",
                              help="If true, do not abort if no labels were found. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--input-file-is-presorted", dest="input_is_presorted",
                              help="If true, the input file is presorted on the column for which values are to be lifted. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--label-file-is-presorted", dest="labels_are_presorted",
                              help="If true, the label file is presorted on the node1 column. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    # TODO: seperate reader_options for the label file.
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_kgtk_file: Path,
        label_kgtk_file: typing.Optional[Path],
        output_kgtk_file: Path,
        node1_column_name: typing.Optional[str],
        label_column_name: typing.Optional[str],
        node2_column_name: typing.Optional[str],
        label_column_value: str,
        lifted_column_suffix: str,
        lift_column_names: typing.List[str],
        remove_label_records: bool = False,
        sort_lifted_labels: bool = True,
        suppress_duplicate_labels: bool = True,
        suppress_empty_columns: bool = False,
        ok_if_no_labels: bool = False,
        input_is_presorted: bool = False,
        labels_are_presorted: bool = False,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % str(input_kgtk_file), file=error_file, flush=True)
        if label_kgtk_file is not None:
            print("-label-file=%s" % label_kgtk_file, file=error_file, flush=True)
        if node1_column_name is not None:
            print("--node1-name=%s" % node1_column_name, file=error_file, flush=True)
        if label_column_name is not None:
            print("--label-name=%s" % label_column_name, file=error_file, flush=True)
        if node2_column_name is not None:
            print("--node2-name=%s" % node2_column_name, file=error_file, flush=True)
        print("--label-value=%s" % label_column_value, file=error_file, flush=True)
        print("--lift-suffix=%s" % lifted_column_suffix, file=error_file, flush=True)
        if lift_column_names is not None and len(lift_column_names) > 0:
            print("--columns-to-lift %s" % " ".join(lift_column_names), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        print("--remove-label-records=%s" % str(remove_label_records))
        print("--sort-lifted-labels=%s" % str(sort_lifted_labels))
        print("--suppress-duplicate-labels=%s" % str(suppress_duplicate_labels))
        print("--suppress-empty-columns=%s" % str(suppress_empty_columns))
        print("--ok-if-no-labels=%s" % str(ok_if_no_labels))
        print("--input-file-is-presorted=%s" % str(input_is_presorted))
        print("--label-file-is-presorted=%s" % str(labels_are_presorted))
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kl: KgtkLift = KgtkLift(
            input_file_path=input_kgtk_file,
            label_file_path=label_kgtk_file,
            node1_column_name=node1_column_name,
            label_column_name=label_column_name,
            node2_column_name=node2_column_name,
            label_column_value=label_column_value,
            lifted_column_suffix=lifted_column_suffix,
            lift_column_names=lift_column_names,
            output_file_path=output_kgtk_file,
            remove_label_records=remove_label_records,
            sort_lifted_labels=sort_lifted_labels,
            suppress_duplicate_labels=suppress_duplicate_labels,
            suppress_empty_columns=suppress_empty_columns,
            ok_if_no_labels=ok_if_no_labels,
            input_is_presorted=input_is_presorted,
            labels_are_presorted=labels_are_presorted,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        kl.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Copy a KGTK file, validating it and producing a clean KGTK file (no
comments, whitespace lines, etc.) as output.

TODO: Need KgtkWriterOptions.

TODO: Need to plumn the infrastructure so we can report at least
a count of how many repair actions took place (per action type).
Ideally, we'ld like the optino to log individual repair actions.

"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Validate a KGTK file and output a clean copy: no comments, whitespace lines, invalid lines, etc. ',
        'description': 'Validate a KGTK file and output a clean copy. ' +
        'Empty lines, whitespace lines, comment lines, and lines with empty required fields are silently skipped. ' +
        'Header errors cause an immediate exception. Data value errors are reported and the line containing them skipped. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert clean_data --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    _expert: bool = parsed_shared_args._expert

    parser.add_argument(      "input_file", nargs="?", help="The KGTK file to read.  May be omitted or '-' for stdin.", type=Path)
    parser.add_argument(      "output_file", nargs="?", help="The KGTK file to write.  May be omitted or '-' for stdout.", type=Path)
    
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, validate_by_default=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: typing.Optional[Path],
        output_file: typing.Optional[Path],
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = False,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,
        **kwargs # Whatever KgtkReaderOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_file) if input_file is not None else "-"), file=error_file)
        print("output: %s" % (str(output_file) if output_file is not None else "-"), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if verbose:
        if input_file is not None:
            print("Cleaning data from '%s'" % str(input_file), file=error_file, flush=True)
        else:
            print ("Cleaning data from stdin", file=error_file, flush=True)
        if output_file is not None:
            print("Writing data to '%s'" % str(output_file), file=error_file, flush=True)
        else:
            print ("Writing data to stdin", file=error_file, flush=True)
                
    try:
        kr: KgtkReader = KgtkReader.open(input_file,
                                         error_file=error_file,
                                         options=reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         output_file,
                                         verbose=verbose, very_verbose=very_verbose)
        
        line_count: int = 0
        row: typing.List[str]
        for row in kr:
            kw.write(row)
            line_count += 1

        kw.close()
        if verbose:
            print("Copied %d clean data lines" % line_count, file=error_file, flush=True)
        return 0

    except Exception as e:
        raise KGTKException(e)




import unittest
from kgtk.cli_entry import cli_entry


class TestEmbedding(unittest.TestCase):

    def test_vector(self):
        # now just test if it is runnning properly or not
        test_input = 'data/embedding_test_input.tsv'
        assert cli_entry("kgtk", "text_embedding", test_input, "--use-cache", "false", "--embedding-projector-metadata-path", "none", "--property-value", "P1629", "P1466") == 0



"""
Join two KGTK edge files or two KGTK node files.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.join.kgtkjoiner import KgtkJoiner
from kgtk.utils.argparsehelpers import optional_bool
from kgtk.value.kgtkvalueoptions import KgtkValueOptions

def parser():
    return {
        'help': 'Join two KGTK files',
        'description': """Join two KGTK edge files or two KGTK node files.

Join keys are extracted from one or both input files and stored in memory,
then the data files are processed in a second pass.  stdin will not work as an
input file if join keys are needed from it.

The output file contains the union of the columns in the two
input files, adjusted for predefined name aliasing.

Specify --left-join to get a left outer join.
Specify --right-join to get a right outer join.
Specify both to get a full outer join (equivalent to cat).
Specify neither to get an inner join.

By default, node files are joined on the id column, while edge files are joined
on the node1 column. The label and node2 columns may be added to the edge file
join criteria.  Alternatively, the left and right file join columns may be
listed explicitly.

To join an edge file to a node file, or to join quasi-KGTK files, use the
following option (enable expert mode for more information):

--mode=NONE

Expert mode provides additional command arguments.
"""
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_argument(      "left_file_path", help="The left-side KGTK file to join (required).", type=Path, default=None)

    parser.add_argument(      "right_file_path", help="The right-side KGTK file to join (required).", type=Path, default=None)

    parser.add_argument(      "--join-on-label", dest="join_on_label",
                              help="If both input files are edge files, include the label column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--join-on-node2", dest="join_on_node2",
                              help="If both input files are edge files, include the node2 column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    
    parser.add_argument(      "--left-file-join-columns", dest="left_join_columns", help="Left file join columns (default=None).", nargs='+')

    parser.add_argument(      "--left-join", dest="left_join", help="Perform a left outer join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument("-o", "--output-file", dest="output_file_path", help="The KGTK file to write (default=%(default)s).", type=Path, default="-")

    parser.add_argument(      "--prefix", dest="prefix",
                              help="An optional prefix applied to right file column names in the output file (default=None).")
    
    parser.add_argument(      "--right-file-join-columns", dest="right_join_columns", help="Right file join columns (default=None).", nargs='+')
    
    parser.add_argument(      "--right-join", dest="right_join", help="Perform a right outer join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--field-separator", dest="field_separator",
                              help=h("Separator for multifield keys (default=%(default)s)")
                              , default=KgtkJoiner.FIELD_SEPARATOR_DEFAULT)

    # Build the command arguments. File arguments can be set for individual
    # files, or for all files.
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="left", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="right", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(left_file_path: typing.Optional[Path],
        right_file_path: typing.Optional[Path],
        left_join: bool,
        right_join: bool,
        join_on_label: bool,
        join_on_node2: bool,
        left_join_columns: typing.Optional[typing.List[str]],
        right_join_columns: typing.Optional[typing.List[str]],
        output_file_path: Path,
        prefix: typing.Optional[str] = None,

        field_separator: str = KgtkJoiner.FIELD_SEPARATOR_DEFAULT,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException


    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    if not right_join:
        if left_file_path is None or str(left_file_path) == "-":
            print("The left file may not be stdin when an inner join or left join is requested.", file=error_file, flush=True)
            return 1

    if not left_join:
        if right_file_path is None or str(right_file_path) == "-":
            print("The right file may not be stdin when an inner join or right join is requested.", file=error_file, flush=True)
            return 1

    if (left_file_path is None or str(left_file_path) == "-") and (right_file_path is None or str(right_file_path) == "-"):
        print("The left and right files may not both be stdin.", file=error_file, flush=True)
        return 1

    if left_file_path is None:
        left_file_path = Path("-")
        
    if right_file_path is None:
        right_file_path = Path("-")
        

    # Build the option structures.
    left_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="left", fallback=True)
    right_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="right", fallback=True)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        # TODO: left_file_path, right_file_path, --join-on-label, etc.
        print("left: %s" % (str(left_file_path) if left_file_path is not None else "-"), file=error_file)
        print("right: %s" % (str(left_file_path) if left_file_path is not None else "-"), file=error_file)
        print("--output-file=%s" % (str(output_file_path) if output_file_path is not None else "-"), file=error_file)
        print("--left-join=%s" % str(left_join), file=error_file)
        print("--right-join=%s" % str(right_join), file=error_file)
        print("--join-on-label=%s" % str(join_on_label), file=error_file)
        print("--join-on-node2=%s" % str(join_on_node2), file=error_file)
        if left_join_columns is not None:
            print("--left-join-columns=%s" % " ".join(left_join_columns), file=error_file)
        if right_join_columns is not None:
            print("--right-join-columns=%s" % " ".join(right_join_columns), file=error_file)
        if prefix is not None:
            print("--prefix=%s" % str(prefix), file=error_file)
        print("--field-separator=%s" % repr(field_separator), file=error_file)
              
        left_reader_options.show(out=error_file, who="left")
        right_reader_options.show(out=error_file, who="right")
        value_options.show(out=error_file)

    try:
        kr: KgtkJoiner = KgtkJoiner(
            left_file_path=left_file_path,
            right_file_path=right_file_path,
            output_path=output_file_path,
            left_join=left_join,
            right_join=right_join,
            join_on_label=join_on_label,
            join_on_node2=join_on_node2,
            left_join_columns=left_join_columns,
            right_join_columns=right_join_columns,
            prefix=prefix,
            field_separator=field_separator,
            left_reader_options=left_reader_options,
            right_reader_options=right_reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        kr.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Import an ntriple file into KGTK file
"""


def parser():
    return {
        'help': 'Import an ntriple file into KGTK file'
    }


def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    parser.add_argument("-i",'--inp', action="store", type=str, dest="input_file",help='input file path')
    parser.add_argument("-o",'--out', action="store", type=str, dest="output_file",help='output file path')
    parser.add_argument("--limit", action="store", type=int, dest="limit",default=None,help='number of lines of input file to run on, default: runs on all')
    
    
    
def run(input_file, output_file, limit):
    # import modules locally
    import re
    import csv
    from kgtk.exceptions import KGTKException
    from kgtk.cli_argparse import KGTKArgumentParser
    
    
    
    try:
        regex = r"\"(?:\\\"|[^\"])+\"|[^\s]+"
        write = True
        errors = 0
        rows = []
        header = []
        header.append('node1')
        header.append('label')
        header.append('node2')
        header.append('node1_prefix')
        header.append('node1_id')
        header.append('label_prefix')
        header.append('label_id')
        header.append('node2_prefix')
        header.append('node2_id')
        header.append('node2_type')
        data_dict = {}
        if write:
            with open(output_file, 'w', newline='') as myfile:
                wr = csv.writer(
                    myfile,
                    quoting=csv.QUOTE_NONE,
                    delimiter="\t",
                    escapechar="\n",
                    quotechar='')
                wr.writerow(header)
        with open(input_file, mode='r') as file:
            print('processing the ntriple file now...')
            for cnt, full_line in enumerate(file):
                matches = re.finditer(regex, full_line, re.MULTILINE)
                line = []
                for m in matches:
                    line.append(m.group())
                keep = True
                if limit and cnt >= limit:
                    break
                if cnt % 500000 == 0 and cnt > 0:
                    print('processed {} lines'.format(cnt))
                if len(line) < 4 or len(line) > 5:
                    errors += 1
                    #print(line)
                    keep = False
                if keep:
                    final_row = []
                    final_row.append(line[0])
                    final_row.append(line[1])
                    subject_isuri = False
                    if line[2].startswith('\"') and line[2].endswith('\"'):
                        line[2] = line[2][1:-1]
                    if line[3] != '.':
                        subject = line[2] + line[3]
                    else:
                        subject = line[2]
                    if '^^' in subject:
                        subject_parts = subject.split('^^')
                        value = subject_parts[0]
                        datatype = subject_parts[1]
                        if 'www.w3.org' in subject:
                            if 'boolean' in datatype:
                                final_value = value.capitalize()
                            elif 'date' in datatype:
                                final_value = '^' + value + 'T00:00:00Z/11'
                            elif 'YearMonth' in datatype:
                                final_value = '^' + value + '-01T00:00:00Z/10'
                            elif 'Year' in datatype:
                                final_value = '^' + value + '-01-01T00:00:00Z/9'
                            else:
                                final_value = value
                        else:
                            datatype = datatype.replace('<', '')
                            datatype = datatype.replace('>', '')
                            final_value = '!' + value + '^^' + datatype
                    else:
                        datatype = ''
                        if subject.startswith('<'):
                            subject_isuri = True
                            final_value = subject
                        else:
                            if '@' in subject:
                                str_parts = subject.split('@')
                                final_value = '\'' + \
                                    str_parts[0].replace("'","\\'") + '\'@' + str_parts[1]
                            else:
                                final_value = '\"' + str(subject) + '\"'
                    final_row.append(final_value)
                    n1_parts = line[0].rsplit('/', 1)
                    final_row += n1_parts
                    label_parts = line[1].rsplit('/', 1)
                    final_row += label_parts
                    if subject_isuri:
                        subject_parts = subject.rsplit('/', 1)
                        final_row += subject_parts
                    else:
                        final_row += [':', final_value]
                    final_row.append(datatype)
                    rows.append(final_row)
                    for i in range(len(final_row)):
                        final_row[i] = final_row[i].replace('<', '')
                        final_row[i] = final_row[i].replace('>', '')
                        if (not subject_isuri) and i == 2:
                            final_row[i] = final_value
                if write:
                    if cnt % 50000 == 0 and cnt > 0:
                        with open(output_file, 'a', newline='') as myfile:
                            for row in rows:
                                wr = csv.writer(
                                    myfile,
                                    quoting=csv.QUOTE_NONE,
                                    delimiter="\t",
                                    escapechar="\n",
                                    quotechar='')
                                wr.writerow(row)
                            rows = []
            if write:
                with open(output_file, 'a', newline='') as myfile:
                    for row in rows:
                        wr = csv.writer(
                            myfile,
                            quoting=csv.QUOTE_NONE,
                            delimiter="\t",
                            escapechar="\n",
                            quotechar='')
                        wr.writerow(row)               
        print('{} invalid lines in ntriple file'.format(errors))
        print('import complete')
    except:
        raise KGTKException



import logging
import re
import redis
import typing
import hashlib
import pandas as pd  # type: ignore
import numpy as np
import io
import math
import pickle
import os
import time

from pyrallel import ParallelProcessor
from collections import defaultdict, OrderedDict
from tqdm import tqdm  # type: ignore
from ast import literal_eval
from sentence_transformers import SentenceTransformer, SentencesDataset, LoggingHandler, losses, models  # type: ignore
from SPARQLWrapper import SPARQLWrapper, JSON, POST, URLENCODED  # type: ignore
from kgtk.exceptions import KGTKException


class EmbeddingVector:
    def __init__(self, model_name=None, query_server=None, cache_config: dict = None, parallel_count=1):
        self._logger = logging.getLogger(__name__)
        if not model_name:
            self.model_name = 'bert-base-nli-mean-tokens'
        # xlnet need to be trained before using, we can't use this for now
        # elif model_name == "xlnet-base-cased":
        #     word_embedding_model = models.XLNet('xlnet-base-cased')
        # # Apply mean pooling to get one fixed sized sentence vector
        #     pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension(),
        #                                pooling_mode_mean_tokens=True,
        #                                pooling_mode_cls_token=False,
        #                                pooling_mode_max_tokens=False)
        #     self.model = SentenceTransformer(modules=[word_embedding_model, pooling_model])
        else:
            self.model_name = model_name
        self._logger.info("Using model {}".format(self.model_name))
        self.model = SentenceTransformer(self.model_name)
        # setup redis cache server
        if query_server is None or query_server == "":
            self.wikidata_server = "https://query.wikidata.org/sparql"
        else:
            self.wikidata_server = query_server
        if cache_config and cache_config.get("use_cache", False):
            host = cache_config.get("host", "dsbox01.isi.edu")
            port = cache_config.get("port", 6379)
            self.redis_server = connect_to_redis(host, port)
        else:
            self.redis_server = None
        self._parallel_count = int(parallel_count)
        self._logger.debug("Running with {} processes.".format(parallel_count))
        self.vectors_map = dict()
        self.node_labels = dict()  # this is used to store {node:label} pairs
        self.candidates = defaultdict(dict)  # this is used to store all node {node:dict()} information
        self.vectors_2D = None
        self.vector_dump_file = None
        self.gt_nodes = set()
        self.metadata = []
        self.gt_indexes = set()
        self.input_format = ""
        self.token_pattern = re.compile(r"(?u)\b\w\w+\b")

    def get_sentences_embedding(self, sentences: typing.List[str], qnodes: typing.List[str]):
        """
            transform a list of sentences to embedding vectors
        """

        if self.redis_server is not None:
            sentence_embeddings = []
            for each_node, each_sentence in zip(qnodes, sentences):
                query_cache_key = each_node + each_sentence
                if self.model_name != "bert-base-wikipedia-sections-mean-tokens":
                    query_cache_key += self.model_name
                cache_res = self.redis_server.get(query_cache_key)
                if cache_res is not None:
                    sentence_embeddings.append(literal_eval(cache_res.decode("utf-8")))
                    # self._logger.error("{} hit!".format(each_node+each_sentence))
                else:
                    each_embedding = self.model.encode([each_sentence], show_progress_bar=False)
                    sentence_embeddings.extend(each_embedding)
                    self.redis_server.set(query_cache_key, str(each_embedding[0].tolist()))
        else:
            sentence_embeddings = self.model.encode(sentences, show_progress_bar=False)
        return sentence_embeddings

    def send_sparql_query(self, query_body: str):
        """
            a simple wrap to send the query and return the returned results
        """
        qm = SPARQLWrapper(self.wikidata_server)
        qm.setReturnFormat(JSON)
        qm.setMethod(POST)
        qm.setRequestMethod(URLENCODED)
        self._logger.debug("Sent query is:")
        self._logger.debug(str(query_body))
        qm.setQuery(query_body)
        try:
            results = qm.query().convert()['results']['bindings']
            return results
        except Exception as e:
            error_message = ("Sending Sparql query to {} failed!".format(self.wikidata_server))
            self._logger.error(error_message)
            self._logger.debug(e, exc_info=True)
            raise KGTKException(error_message)

    def _get_labels(self, nodes: typing.List[str]):
        nodes_need_query = set()
        for each in nodes:
            if each not in self.node_labels:
                nodes_need_query.add(each)
        if nodes_need_query:
            query_nodes = " ".join(["wd:{}".format(each) for each in nodes_need_query])
            query = """
            select ?item ?nodeLabel
            where { 
              values ?item {""" + query_nodes + """}
              ?item rdfs:label ?nodeLabel.
              FILTER(LANG(?nodeLabel) = "en").
            }
            """
            results2 = self.send_sparql_query(query)
            for each_res in results2:
                node_id = each_res['item']['value'].split("/")[-1]
                nodes_need_query.remove(node_id)
                value = each_res['nodeLabel']['value']
                self.node_labels[node_id] = value

            # for those nodes we can't find label, just add this to dict to prevent query again
            for each_node in nodes_need_query:
                self.node_labels[each_node] = each_node

    def _get_labels_and_descriptions(self, query_qnodes: str, need_find_label: bool, need_find_description: bool):
        query_body = """
            select ?item ?itemDescription ?itemLabel
            where {
              values ?item {""" + query_qnodes + """ }
                 SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
            }
        """
        results = self.send_sparql_query(query_body)
        for each in results:
            each_node = each['item']['value'].split("/")[-1]
            if 'itemDescription' in each:
                description = each['itemDescription']['value']
            else:
                description = ""
            if "itemLabel" in each:
                label = each['itemLabel']['value']
            else:
                label = ""
            if need_find_label:
                self.candidates[each_node]["label_properties"] = [label]
            if need_find_description:
                self.candidates[each_node]["description_properties"] = [description]

    def _get_property_values(self, query_qnodes: str, properties: dict, properties_reversed: dict):
        """
        run sparql query to get corresponding property values of given q nodes
        """
        used_p_node_ids = set()
        all_needed_properties = ""
        for part_name, part in properties.items():
            if part_name == "isa_properties":
                self._get_labels(part)

        for each_node, role in properties_reversed.items():
            if role != {"has_properties"} and each_node not in {"label", "description", "all"}:
                all_needed_properties += "wdt:{} ".format(each_node)

        query_body = """
        select ?item ?properties ?eachPropertyValueLabel
        where {{
          values ?item {{{all_nodes}}}
          values ?properties {{{properties}}}
          ?item ?properties ?eachPropertyValue.
          SERVICE wikibase:label {{ bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }}
        }}
        """.format(all_nodes=query_qnodes, properties=all_needed_properties)
        results = self.send_sparql_query(query_body)

        for each_res in results:
            node_id = each_res['item']['value'].split("/")[-1]
            node_property = each_res['properties']['value'].split("/")[-1]
            roles = properties_reversed[node_property]
            value = each_res['eachPropertyValueLabel']['value']
            if node_property in properties["isa_properties"] and self.node_labels[node_property].endswith("of"):
                value = self.node_labels[node_property] + "||" + value
            used_p_node_ids.add(node_property)
            for each_role in roles:
                if each_role != "property_values":
                    if each_role in self.candidates[node_id]:
                        self.candidates[node_id][each_role].add(value)
                    else:
                        self.candidates[node_id][each_role] = {value}
        return used_p_node_ids

    def _get_all_properties(self, query_qnodes: str, used_p_node_ids: set, properties: dict):
        """
        run sparql query to get all properties of given q nodes
        """
        has_properties_set = set(properties["has_properties"])
        query_body3 = """
            select DISTINCT ?item ?p_entity ?p_entityLabel
            where {
              values ?item {""" + query_qnodes + """}
              ?item ?p ?o.
              FILTER regex(str(?p), "^http://www.wikidata.org/prop/P", "i")
              BIND (IRI(REPLACE(STR(?p), "http://www.wikidata.org/prop", "http://www.wikidata.org/entity")) AS ?p_entity) .
              SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
            }
            """
        results3 = self.send_sparql_query(query_body3)
        for each in results3:
            node_name = each['item']['value'].split("/")[-1]
            p_node_id = each['p_entity']['value'].split("/")[-1]
            p_node_label = each['p_entityLabel']['value']
            if p_node_id not in used_p_node_ids:
                if has_properties_set == {"all"} or p_node_id in has_properties_set:
                    if "has_properties" in self.candidates[node_name]:
                        self.candidates[node_name]["has_properties"].add(p_node_label)
                    else:
                        self.candidates[node_name]["has_properties"] = {p_node_label}

    def get_item_description(self, target_properties: dict, properties_reversed: dict,
                             qnodes: typing.Union[set, typing.List[str]]):
        """
            use sparql query to get the descriptions of given Q nodes
        """
        # find_all_properties = False
        if "all" in properties_reversed:
            # find_all_properties = True
            _ = properties_reversed.pop("all")
        self._logger.info("Need to find all properties.")

        hash_generator = hashlib.md5()
        # sort to ensure the hash key same
        target_properties = OrderedDict(sorted(target_properties.items()))
        hash_generator.update(str(target_properties).encode('utf-8'))
        properties_list_hash = "||" + str(hash_generator.hexdigest())

        sentences_cache_dict = {}
        if self.redis_server is not None:
            for each_node in qnodes:
                cache_key = each_node + properties_list_hash
                cache_res = self.redis_server.get(cache_key)
                self._logger.debug("Cached key is: {}".format(cache_key))
                if cache_res is not None:
                    self._logger.debug("Cache hit {}".format(cache_key))
                    sentences_cache_dict[each_node] = cache_res.decode("utf-8")

        self._logger.debug("Cached for those nodes {} / {}".format(len(sentences_cache_dict), len(qnodes)))
        self._logger.debug(str(set(sentences_cache_dict.keys())))
        self._logger.debug("Need run query for those nodes {} / {}:".format(len(qnodes) - len(sentences_cache_dict), len(qnodes)))

        # we do not need to get those node again
        if len(sentences_cache_dict) > 0:
            qnodes = set(qnodes) - set(sentences_cache_dict.keys())
        self._logger.debug(str(qnodes))

        # only need to do query when we still have remained nodes
        if len(qnodes) > 0:
            query_qnodes = ""
            for each in qnodes:
                query_qnodes += "wd:{} ".format(each)

            need_find_label = "label" in target_properties["label_properties"]
            need_find_description = "description" in target_properties["description_properties"]
            # this is used to get corresponding labels / descriptions
            if need_find_label or need_find_description:
                self._get_labels_and_descriptions(query_qnodes, need_find_label, need_find_description)

            # this is used to get corresponding labels of properties values
            used_p_node_ids = self._get_property_values(query_qnodes, target_properties, properties_reversed)

            # if need get all properties, we need to run extra query
            # if find_all_properties:
            self._get_all_properties(query_qnodes, used_p_node_ids, target_properties)

        for each_node_id in qnodes:
            each_sentence = self.attribute_to_sentence(self.candidates[each_node_id], each_node_id)
            self.candidates[each_node_id]["sentence"] = each_sentence
            # add to cache
            if self.redis_server is not None:
                response = self.redis_server.set(each_node_id + properties_list_hash, each_sentence)
                if response:
                    self._logger.debug("Pushed cache for {} success.".format(each_node_id + properties_list_hash))

        for each_node_id, sentence in sentences_cache_dict.items():
            self.candidates[each_node_id]["sentence"] = sentence

    def _process_one(self, args):
        """
        one process for multiprocess calling, should not be used for any other function
        :param args: args to receive from main process
        :return: corresponding node vector and attribute
        """
        node_id = args["node_id"]
        each_node_attributes = args["attribute"]
        concat_sentence = self.attribute_to_sentence(each_node_attributes, node_id)
        vectors = self.get_sentences_embedding([concat_sentence], [node_id])[0]
        return {"v_" + node_id: vectors, "c_" + node_id: each_node_attributes}

    def _multiprocess_collector(self, data):
        for k, v in data.items():
            if k.startswith("v_"):
                k = k.replace("v_", "")
                self.vectors_map[k] = v
            else:
                k = k.replace("c_", "")
                self.candidates[k] = v

    def read_input(self, input_file: io.TextIOWrapper, target_properties: dict, property_labels_dict: dict,
                   skip_nodes_set: set = None, input_format: str = "kgtk_format",
                   black_list_set: typing.Optional[set] = None
                   ):
        """
            load the input candidates files
        """
        self.node_labels.update(property_labels_dict)
        # reverse sentence property to be {property : role)
        properties_reversed = defaultdict(set)
        for k, v in target_properties.items():
            for each_property in v:
                properties_reversed[each_property].add(k)

        if input_format == "test_format":
            self.input_format = input_format
            input_df = pd.read_csv(input_file, dtype=object)
            gt = {}
            count = 0
            if "GT_kg_id" in input_df.columns:
                gt_column_id = "GT_kg_id"
            elif "kg_id" in input_df.columns:
                gt_column_id = "kg_id"
            else:
                raise KGTKException("Can't find ground truth id column! It should either named as `GT_kg_id` or `kg_id`")

            for _, each in input_df.iterrows():
                temp = []
                if isinstance(each["candidates"], str):
                    temp = str(each['candidates']).split("|")
                elif each['candidates'] is np.nan or math.isnan(each['candidates']):
                    temp = []

                to_remove_q = set()
                if each[gt_column_id] is np.nan:
                    self._logger.warning("Ignore NaN gt value form {}".format(str(each)))
                    each[gt_column_id] = ""
                gt_nodes = each[gt_column_id].split(" ")
                label = str(each["label"])
                if len(gt_nodes) == 0:
                    self._logger.error("Skip a row with no ground truth node given: as {}".format(str(each)))
                    continue
                if label == "":
                    self._logger.error("Skip a row with no label given: as {}".format(str(each)))
                    continue
                temp.extend(gt_nodes)

                for each_q in temp:
                    if skip_nodes_set is not None and each_q in skip_nodes_set:
                        to_remove_q.add(each_q)
                temp = set(temp) - to_remove_q
                count += len(temp)
                self.gt_nodes.add(each[gt_column_id])
                self.get_item_description(target_properties, properties_reversed, temp)

            self._logger.info("Totally {} rows with {} candidates loaded.".format(str(len(gt)), str(count)))

        elif input_format == "kgtk_format":
            # assume the input edge file is sorted
            if "all" in properties_reversed:
                _ = properties_reversed.pop("all")
                add_all_properties = True
            else:
                add_all_properties = False

            self.input_format = input_format

            # get header
            headers = input_file.readline().replace("\n", "").split("\t")
            if len(headers) < 3:
                raise KGTKException(
                    "No enough columns found on given input file. Only {} columns given but at least 3 needed.".format(
                        len(headers)))
            elif "node" in headers and "property" in headers and "value" in headers:
                column_references = {"node": headers.index("node"),
                                     "property": headers.index("property"),
                                     "value": headers.index("value")}

            elif "node1" in headers and "label" in headers and "node2" in headers:
                column_references = {"node": headers.index("node1"),
                                     "property": headers.index("label"),
                                     "value": headers.index("node2")}

            elif len(headers) == 3:
                column_references = {"node": 0,
                                     "property": 1,
                                     "value": 2}
            else:
                missing_column1 = {"node", "property", "value"} - set(headers)
                missing_column2 = {"node1", "label", "node2"} - set(headers)
                missing_column = missing_column1 if len(missing_column1) < len(missing_column2) else missing_column2
                raise KGTKException("Missing column: {}".format(missing_column))

            self._logger.debug("column index information: ")
            self._logger.debug(str(column_references))
            # read contents
            each_node_attributes = {"has_properties": [], "isa_properties": [], "label_properties": [],
                                    "description_properties": [], "has_properties_values": []}
            current_process_node_id = None

            if self._parallel_count > 1:
                # need to set with spawn mode to initialize with multiple cuda in multiprocess
                from multiprocessing import set_start_method
                set_start_method('spawn')
                pp = ParallelProcessor(self._parallel_count, self._process_one, collector=self._multiprocess_collector)
                pp.start()

            for each_line in input_file:
                each_line = each_line.replace("\n", "").split("\t")
                node_id = each_line[column_references["node"]]
                # skip nodes id in black list
                if black_list_set and node_id in black_list_set:
                    continue

                node_property = each_line[column_references["property"]]
                node_value = each_line[column_references["value"]]
                # remove @ mark
                if "@" in node_value and node_value[0] != "@":
                    node_value = node_value[:node_value.index("@")]

                # in case we meet an empty value, skip it
                if node_value == "":
                    self._logger.warning("""Skip line "{}" because of empty value.""".format(each_line))
                    continue

                # remove extra double quote " and single quote '
                while len(node_value) >= 3 and node_value[0] == '"' and node_value[-1] == '"':
                    node_value = node_value[1:-1]
                while len(node_value) >= 3 and node_value[0] == "'" and node_value[-1] == "'":
                    node_value = node_value[1:-1]

                if current_process_node_id != node_id:
                    if current_process_node_id is None:
                        current_process_node_id = node_id
                    else:
                        # if we get to next id, concat all properties into one sentence to represent the Q node

                        # for multi process
                        if self._parallel_count > 1:
                            each_arg = {"node_id": current_process_node_id, "attribute": each_node_attributes}
                            pp.add_task(each_arg)
                        # for single process
                        else:
                            concat_sentence = self.attribute_to_sentence(each_node_attributes, current_process_node_id)
                            each_node_attributes["sentence"] = concat_sentence
                            self.candidates[current_process_node_id] = each_node_attributes

                        # after write down finish, we can clear and start parsing next one
                        each_node_attributes = {"has_properties": [], "isa_properties": [], "label_properties": [],
                                                "description_properties": [], "has_properties_values": []}
                        # update to new id
                        current_process_node_id = node_id

                if node_property in properties_reversed:
                    roles = properties_reversed[node_property].copy()
                    node_value = self.get_real_label_name(node_value)
                    # if we get property_values, it should be saved to isa-properties part
                    if "property_values" in roles:
                        # for property values part, changed to be "{property} {value}"
                        node_value_combine = self.get_real_label_name(node_property) + " " + self.get_real_label_name(node_value)
                        each_node_attributes["has_properties_values"].append(node_value_combine)
                        # remove those 2 roles in case we have duplicate using of this node later
                        roles.discard("property_values")
                        roles.discard("has_properties")
                    for each_role in roles:
                        each_node_attributes[each_role].append(node_value)
                elif add_all_properties:  # add remained properties if need all properties
                    each_node_attributes["has_properties"].append(self.get_real_label_name(node_property))

                # close multiprocess pool
                if self._parallel_count > 1:
                    pp.task_done()
                    pp.join()
        else:
            raise KGTKException("Unknown input format {}".format(input_format))

        self._logger.info("Totally {} Q nodes loaded.".format(len(self.candidates)))
        try:
            file_path = input_file.name
            file_name = file_path[:file_path.rfind(".")]
        except AttributeError:
            file_name = "input_from_memory"
        self.vector_dump_file = "dump_vectors_{}_{}.pkl".format(file_name, self.model_name)
        # self._logger.debug("The cache file name will be {}".format(self.vector_dump_file))

    def get_real_label_name(self, node):
        if node in self.node_labels:
            return self.node_labels[node]
        else:
            return node

    def attribute_to_sentence(self, attribute_dict: dict, node_id=None):
        concated_sentence = ""
        have_isa_properties = False
        # sort the properties to ensure the sentence always same
        attribute_dict = {key: sorted(list(value)) for key, value in attribute_dict.items() if len(value) > 0}
        if "label_properties" in attribute_dict and len(attribute_dict["label_properties"]) > 0:
            concated_sentence += self.get_real_label_name(attribute_dict["label_properties"][0])
        if "description_properties" in attribute_dict and len(attribute_dict["description_properties"]) > 0:
            if concated_sentence != "" and attribute_dict["description_properties"][0] != "":
                concated_sentence += ", "
            concated_sentence += self.get_real_label_name(attribute_dict["description_properties"][0])
        if "isa_properties" in attribute_dict and len(attribute_dict["isa_properties"]) > 0:
            have_isa_properties = True
            temp = ""
            for each in attribute_dict["isa_properties"]:
                each = self.get_real_label_name(each)
                if "||" in each:
                    if "instance of" in each:
                        each = each.split("||")[1]
                    else:
                        each = each.replace("||", " ")
                temp += each + ", "
            if concated_sentence != "" and temp != "":
                concated_sentence += " is "
            elif concated_sentence == "":
                concated_sentence += "It is "
            # remove last ", "
            concated_sentence += temp[:-2]
        if "has_properties_values" in attribute_dict and len(attribute_dict["has_properties_values"]) > 0:
            temp = [self.get_real_label_name(each) for each in attribute_dict["has_properties_values"]]
            if concated_sentence != "":
                if not have_isa_properties:
                    concated_sentence += " "
                else:
                    concated_sentence += ", "
            else:
                concated_sentence += "It "
            concated_sentence += " and ".join(temp)
        if "has_properties" in attribute_dict and len(attribute_dict["has_properties"]) > 0:
            temp = [self.get_real_label_name(each) for each in attribute_dict["has_properties"]]
            if concated_sentence != "" and temp[0] != "":
                if have_isa_properties:
                    concated_sentence += ", and has "
                else:
                    concated_sentence += " has "
            elif temp[0] != "":
                concated_sentence += "It has "
            concated_sentence += " and ".join(temp)
        # add ending period
        if concated_sentence != "":
            concated_sentence += "."
        self._logger.debug("Transform node {} --> {}".format(node_id, concated_sentence))
        return concated_sentence

    def get_vectors(self):
        """
            main function to get the vector representations of the descriptions
        """
        if self._parallel_count == 1:
            start_all = time.time()
            self._logger.info("Now generating embedding vector.")
            for q_node, each_item in tqdm(self.candidates.items()):
                # do process for each row(one target)
                sentence = each_item["sentence"]
                if isinstance(sentence, bytes):
                    sentence = sentence.decode("utf-8")
                vectors = self.get_sentences_embedding([sentence], [q_node])
                self.vectors_map[q_node] = vectors[0]
            self._logger.info("Totally used {} seconds.".format(str(time.time() - start_all)))
        else:
            # Skip get vector function because we already get them
            pass

    def dump_vectors(self, file_name, type_=None):
        if file_name.endswith(".pkl"):
            file_name = file_name.replace(".pkl", "")
        if type_ == "2D":
            with open(file_name + ".pkl", "wb") as f:
                pickle.dump(self.vectors_2D, f)
            dimension = len(self.vectors_2D[0])
            with open(file_name + ".tsv", "w") as f:
                for each in self.vectors_2D:
                    for i, each_val in enumerate(each):
                        _ = f.write(str(each_val))
                        if i != dimension - 1:
                            _ = f.write("\t")
                    _ = f.write("\n")
        elif type_ == "metadata":
            with open(file_name + "_metadata.tsv", "w") as f:
                for each in self.metadata:
                    _ = f.write(each + "\n")
        else:
            with open(file_name + ".pkl", "wb") as f:
                pickle.dump(self.vectors_map, f)
            with open(file_name + ".tsv", "w") as f:
                for each in self.vectors_map.values():
                    for i in each:
                        _ = f.write(str(i) + "\t")
                    _ = f.write("\n")

    def print_vector(self, vectors, output_properties: str = "text_embedding",
                     output_format="kgtk_format", save_embedding_sentence=False):
        self._logger.debug("START printing the vectors")
        if output_format == "kgtk_format":
            print("node\tproperty\tvalue\n", end="")
            all_nodes = list(self.vectors_map.keys())
            ten_percent_len = math.ceil(len(vectors) / 10)
            for i, each_vector in enumerate(vectors):
                if i % ten_percent_len == 0:
                    percent = i / ten_percent_len * 10
                    self._logger.debug("Finished {}%".format(percent))
                print("{}\t{}\t".format(all_nodes[i], output_properties), end="")
                for each_dimension in each_vector[:-1]:
                    print(str(each_dimension) + ",", end="")
                print(str(each_vector[-1]))
                if save_embedding_sentence:
                    print("{}\t{}\t{}".format(all_nodes[i], "embedding_sentence", self.candidates[all_nodes[i]]["sentence"]))

        elif output_format == "tsv_format":
            for each_vector in vectors:
                for each_dimension in each_vector[:-1]:
                    print(str(each_dimension) + "\t", end="")
                print(str(each_vector[-1]))
        self._logger.debug("END printing the vectors")

    def plot_result(self, output_properties: dict, input_format="kgtk_format",
                    output_uri: str = "", output_format="kgtk_format",
                    dimensional_reduction="none", dimension_val=2,
                    save_embedding_sentence=False
                    ):
        """
            transfer the vectors to lower dimension so that we can plot
            Then save the 2D vector file for further purpose
        """
        self.vectors_map = {k: v for k, v in sorted(self.vectors_map.items(), key=lambda item: item[0], reverse=True)}
        vectors = list(self.vectors_map.values())
        # reduce dimension if needed
        if dimensional_reduction.lower() == "tsne":
            self._logger.warning("Start running TSNE to reduce dimension. It will take some time.")
            start = time.time()
            from sklearn.manifold import TSNE  # type: ignore
            self.vectors_2D = TSNE(n_components=int(dimension_val), random_state=0).fit_transform(vectors)
            self._logger.info("Totally used {} seconds.".format(time.time() - start))
        elif dimensional_reduction.lower() == "pca":
            self._logger.warning("Start running PCA to reduce dimension. It will take some time.")
            start = time.time()
            from sklearn.decomposition import PCA  # type: ignore
            self.vectors_2D = PCA(n_components=int(dimension_val)).fit_transform(vectors)
            self._logger.info("Totally used {} seconds.".format(time.time() - start))
        elif dimensional_reduction.lower() == "none":
            self._logger.info("Not run dimensional reduction algorithm.")
        else:
            raise KGTKException("Unknown or unsupport dimensional reduction type: {}".format(dimensional_reduction))

        if output_uri not in {"", "none"}:
            if not os.path.exists(output_uri):
                raise ValueError("The given metadata output folder does not exist!")

            metadata_output_path = os.path.join(output_uri, self.vector_dump_file.split("/")[-1])
            if input_format == "test_format":
                gt_indexes = set()
                vector_map_keys = list(self.vectors_map.keys())
                for each_node in self.gt_nodes:
                    gt_indexes.add(vector_map_keys.index(each_node))

                self.metadata.append("Q_nodes\tType\tLabel\tDescription")
                for i, each in enumerate(self.vectors_map.keys()):
                    label = self.node_labels[each]
                    description = self.candidates[each]["sentence"]
                    if i in gt_indexes:
                        self.metadata.append("{}\tground_truth_node\t{}\t{}".format(each, label, description))
                    else:
                        self.metadata.append("{}\tcandidates\t{}\t{}".format(each, label, description))
                self.gt_indexes = gt_indexes

            elif input_format == "kgtk_format":
                if len(output_properties.get("metadata_properties", [])) == 0:
                    for k, v in self.candidates.items():
                        label = v.get("label_properties", "")
                        if len(label) > 0 and isinstance(label, list):
                            label = label[0]
                        description = v.get("description_properties", "")
                        if len(description) > 0 and isinstance(description, list):
                            description = description[0]
                        self.metadata.append("{}\t\t{}\t{}".format(k, label, description))
                else:
                    required_properties = output_properties["metadata_properties"]
                    self.metadata.append("node\t" + "\t".join(required_properties))
                    for k, v in self.candidates.items():
                        each_metadata = k + "\t"
                        for each in required_properties:
                            each_metadata += v.get(each, " ") + "\t"
                        self.metadata.append(each_metadata)
            self.dump_vectors(metadata_output_path, "metadata")

        if self.vectors_2D is not None:
            self.print_vector(self.vectors_2D, output_properties.get("output_properties"), output_format, save_embedding_sentence)
        else:
            self.print_vector(vectors, output_properties.get("output_properties"), output_format, save_embedding_sentence)

    def evaluate_result(self):
        """
        for the ground truth nodes, evaluate the average distance to the centroid, the lower the average distance,
        the better clustering results should be
        """
        centroid = None
        gt_nodes_vectors = []
        if len(self.gt_indexes) == 0:
            points = set(range(len(self.vectors_map)))
        else:
            points = self.gt_indexes
        for i, each in enumerate(self.vectors_map.keys()):
            if i in points:
                if centroid is None:
                    centroid = np.array(self.vectors_map[each])
                else:
                    centroid += np.array(self.vectors_map[each])
                gt_nodes_vectors.append(self.vectors_map[each])
        centroid = centroid / len(points)

        distance_sum = 0
        for each in gt_nodes_vectors:
            distance_sum += self.calculate_distance(each, centroid)
        self._logger.info("The average distance for the ground truth nodes to centroid is {}".format(distance_sum / len(points)))

    @staticmethod
    def calculate_distance(a, b):
        if len(a) != len(b):
            raise KGTKException("Vector dimension are different!")
        dist = 0
        for v1, v2 in zip(a, b):
            dist += (v1 - v2) ** 2
        dist = dist ** 0.5
        return dist


def connect_to_redis(host, port):
    _logger = logging.getLogger(__name__)
    redis_server = redis.Redis(host=host, port=port, db=0)
    try:
        _ = redis_server.get("foo")
        _logger.debug("Cache server {}:{} connected!".format(host, port))
    except Exception as e:
        _logger.error("Cache server {}:{} is not able to be connected! Will not use cache!".format(host, port))
        _logger.debug(e, exc_info=True)
        redis_server = None
    return redis_server



import sys
import importlib
import pkgutil
import itertools

from kgtk import cli
from kgtk.exceptions import KGTKExceptionHandler, KGTKArgumentParseException
from kgtk import __version__
from kgtk.cli_argparse import KGTKArgumentParser, add_shared_arguments, add_default_arguments
import sh # type: ignore


# module name should NOT start with '__' (double underscore)
handlers = [x.name for x in pkgutil.iter_modules(cli.__path__)
                   if not x.name.startswith('__')]

# import signal
# signal.signal(signal.SIGPIPE, signal.SIG_DFL)

pipe_delimiter = '/'
ret_code = 0


def cmd_done(cmd, success, exit_code):
    # cmd.cmd -> complete command line
    global ret_code
    ret_code = exit_code


def cli_entry(*args):
    """
    Usage:
        kgtk <command> [options]
    """
    global ret_code

    # get all arguments
    if not args:
        args = tuple(sys.argv)
    args = args[1:]

    # base parser for shared arguments
    base_parser = KGTKArgumentParser(add_help=False)
    base_parser.add_argument(
        '-V', '--version',
        action='version',
        version='KGTK %s' % __version__,
        help='show KGTK version number and exit.'
    )
    shared_args = base_parser.add_argument_group('shared optional arguments')
    shared_args.add_argument('--debug', dest='_debug', action='store_true', default=False, help='enable debug mode')
    shared_args.add_argument('--expert', dest='_expert', action='store_true', default=False, help='enable expert mode')
    add_shared_arguments(shared_args)

    # parse shared arguments
    parsed_shared_args, rest_args = base_parser.parse_known_args(args)
    shared_args = tuple(filter(lambda a: a not in rest_args, args))
    args = tuple(rest_args)

    # complete parser, load sub-parser of each module
    parser = KGTKArgumentParser(
        parents=[base_parser], prog='kgtk',
        description='kgtk --- Knowledge Graph Toolkit',
    )
    sub_parsers = parser.add_subparsers(
        metavar='command',
        dest='cmd'
    )
    subparser_lookup = {}
    sub_parsers.required = True
    for h in handlers:
        mod = importlib.import_module('.{}'.format(h), 'kgtk.cli')
        subp = mod.parser()
        # only create sub-parser with sub-command name and defer full build
        sub_parser = sub_parsers.add_parser(h, **subp)
        subparser_lookup[h] = (mod, sub_parser)
        if 'aliases' in subp:
            for alias in subp['aliases']:
                subparser_lookup[alias] = (mod, sub_parser)

    # add root level usage after sub-parsers are created
    # this won't pollute help info in sub-parsers
    parser.usage = '%(prog)s [options] command [ / command]*'

    # parse internal pipe
    pipe = [tuple(y) for x, y in itertools.groupby(args, lambda a: a == pipe_delimiter) if not x]
    if len(pipe) == 0:
        parser.print_usage()
        parser.exit(KGTKArgumentParseException.return_code)
    elif len(pipe) == 1:  # single command
        cmd_args = pipe[0]

        cmd_name = cmd_args[0]
        # build sub-parser
        if cmd_name in subparser_lookup:
            mod, sub_parser = subparser_lookup[cmd_name]
            add_default_arguments(sub_parser)  # call this before adding other arguments
            if hasattr(mod, 'add_arguments_extended'):
                mod.add_arguments_extended(sub_parser, parsed_shared_args)
            else:
                mod.add_arguments(sub_parser)
        parsed_args = parser.parse_args(cmd_args)

        # load module
        kwargs = {}
        func = None
        if parsed_args.cmd:
            h = parsed_args.cmd
            func = mod.run

            # remove sub-command name
            kwargs = vars(parsed_args)
            del kwargs['cmd']

            # set shared arguments
            for sa in vars(parsed_shared_args):
                if sa not in sub_parsers.choices[h].shared_arguments:
                    del kwargs[sa]
                else:
                    kwargs[sa] = getattr(parsed_shared_args, sa)

        # run module
        kgtk_exception_handler = KGTKExceptionHandler(debug=parsed_shared_args._debug)
        ret_code = kgtk_exception_handler(func, **kwargs)
    else:  # piped commands
        concat_cmd_str = None
        for idx, cmd_args in enumerate(pipe):
            # add shared arguments
            cmd_str = ', '.join(['"{}"'.format(a) for a in shared_args])
            if cmd_str:
                cmd_str += ', '
            # parse command and options
            cmd_str += ', '.join(['"{}"'.format(a) for a in cmd_args])
            # add other common arguments
            cmd_str += ', _bg_exc=False, _done=cmd_done'  # , _err=sys.stdout
            # add specific arguments
            if idx == 0:  # first command
                concat_cmd_str = 'sh.kgtk({}, _in=sys.stdin, _piped=True)'.format(cmd_str)
            elif idx + 1 == len(pipe):  # last command
                concat_cmd_str = 'sh.kgtk({}, {}, _out=sys.stdout)'.format(concat_cmd_str, cmd_str)
            else:
                concat_cmd_str = 'sh.kgtk({}, {}, _piped=True)'.format(concat_cmd_str, cmd_str)
        try:
            process = eval(concat_cmd_str)
            process.wait()
        except sh.SignalException_SIGPIPE:
            pass
        except sh.ErrorReturnCode as e:
            # mimic parser exit
            parser.exit(KGTKArgumentParseException.return_code, e.stderr.decode('utf-8'))

    return ret_code




import csv
from graph_tool import load_graph_from_csv
from graph_tool.util import find_edge
from graph_tool.topology import label_components
from kgtk.kgtkformat import KgtkFormat

import attr
from pathlib import Path
import sys
import typing
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions

from kgtk.value.kgtkvalueoptions import KgtkValueOptions


@attr.s(slots=True, frozen=True)
class ConnectedComponents(KgtkFormat):
    input_file_path: typing.Optional[Path] = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(Path)))

    output_file_path: typing.Optional[Path] = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(Path)), default=None)
    no_header: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    undirected: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    strong: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    properties: str = attr.ib(validator=attr.validators.instance_of(str), default='')

    input_reader_options: typing.Optional[KgtkReaderOptions] = attr.ib(default=None)
    filter_reader_options: typing.Optional[KgtkReaderOptions] = attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def get_edge_key_columns(self, kr: KgtkReader, who: str) -> typing.List[int]:
        if kr.node1_column_idx < 0:
            raise ValueError("The node1 column is missing from the %s edge file." % who)
        if kr.label_column_idx < 0:
            raise ValueError("The label column is missing from the %s edge file." % who)
        if kr.node2_column_idx < 0:
            raise ValueError("The node2 column is missing from the %s edge file." % who)
        return [kr.node1_column_idx, kr.label_column_idx, kr.node2_column_idx]

    def get_key_columns(self, kr: KgtkReader, who: str) -> typing.List[int]:

        if not (kr.is_node_file or kr.is_edge_file):
            raise ValueError("The %s file is a quasi-KGTK file.  Please supply its keys." % who)

        return self.get_edge_key_columns(kr, who)

    def process(self):
        input_kr: KgtkReader = KgtkReader.open(self.input_file_path,
                                               error_file=self.error_file,
                                               who="input",
                                               options=self.input_reader_options,
                                               value_options=self.value_options,
                                               verbose=self.verbose,
                                               very_verbose=self.very_verbose,
                                               )

        input_key_columns: typing.List[int] = self.get_key_columns(input_kr, "input")
        label_col_idx = input_key_columns[1]
        label = '{}{}'.format('c', label_col_idx)

        g = load_graph_from_csv(str(self.input_file_path.absolute()), not (self.undirected),
                                skip_first=not (self.no_header),
                                hashed=True,
                                csv_options={'delimiter': '\t'}, ecols=(input_key_columns[0], input_key_columns[2]))

        es = []
        header = ['node1', 'label', 'node2']
        if self.properties:
            properties = self.properties.split(',')
            for e in properties:
                es += (find_edge(g, g.edge_properties[label], e))
            g.clear_edges()
            g.add_edge_list(list(set(es)))
        comp, hist = label_components(g, directed=self.strong)
        if self.output_file_path:
            f = open(self.output_file_path, 'w')
            wr = csv.writer(f, quoting=csv.QUOTE_NONE, delimiter="\t", escapechar="\n", quotechar='')
            wr.writerow(header)
            for v, c in enumerate(comp):
                wr.writerow([g.vertex_properties['name'][v], 'connected_component', c])
            f.close()
        else:
            sys.stdout.write('%s\t%s\t%s\n' % ('node1', 'label', 'node2'))
            for v, c in enumerate(comp):
                sys.stdout.write('%s\t%s\t%s\n' % (g.vertex_properties['name'][v], 'connected_component', str(c)))
